require 'envoi/mam/agent'
require 'envoi/mam/agent/transfer_client/aspera'
require 'envoi/mam/agent/transfer_client/s3'

require 'envoi/restore/sqs-queue-worker'
require 'envoi/restore/glacier-restore-event-handler'

module Envoi

  class Restore

    class SQSMessageHandler

      attr_accessor :logger
      attr_accessor :queue_handler, :event_handler, :message
      attr_accessor :worker_thread

      def initialize(args = {})
        @logger = args[:logger]
        @queue_handler = args[:queue_handler]
        @message = args[:message]
        @event_handler = args[:event_handler]
      end

      def process
        msg_body_parsed = JSON.parse(message.body)
        event = msg_body_parsed

        result = nil
        worker_thread = Thread.new(event, @event_handler, result) do
          result = @event_handler.process_event(event)
        end

        visibility_timeout = 5
        new_visibility_timeout = 60

        while worker_thread.alive?
          change_message_visibility_timeout(new_visibility_timeout)
          sleep visibility_timeout
        end
        if result && result[:success]
          delete_message
        end
      end

      def change_message_visibility_timeout(visibility_timeout)
        logger.debug { "Updating Visibility Timeout for Message #{message.message_id}..."}
        queue_handler.poller.change_message_visibility_timeout(message, visibility_timeout)
      end

      def delete_message
        logger.debug { "Deleting Message #{message.message_id}..."}
        queue_handler.poller.delete_message(message)
      end

    end

    class Agent < Envoi::Mam::Agent

      DEFAULT_ASPERA_ARGS = '-v -k3 --overwrite=diff -P 33001'
      DEFAULT_DESTINATION_PATH = '.'
      DEFAULT_THREAD_LIMIT = 10

      attr_accessor :default_aspera_ascp_args,
                    :default_aspera_ascp_path

      def after_initialize
        args = initial_args
        @running = false
        @should_stop = false
        @default_aspera_ascp_path   = args[:default_aspera_ascp_path]
        @default_aspera_args = args[:default_ascp_args] || DEFAULT_ASPERA_ARGS

        @thread_limit = args.fetch(:thread_limit, DEFAULT_THREAD_LIMIT)

        initialize_queue_handler
        initialize_event_handler
      end

      def system_config
        @system_config ||= config['restore'] || { }
      end

      def sqs_config
        @sqs_config ||= system_config['sqs'] || { skip_delete: true}
      end

      def initialize_queue_handler
        @queue_handler = SQSQueueWorker.new(config: sqs_config)
        @queue_handler.poller.before_request do |stats|
          throw :stop_polling if @should_stop
        end
      end

      def initialize_event_handler
        @event_handler = GlacierRestoreEventHandler.new(config: sqs_config, logger: logger)
      end

      def run
        @running = true
        logger.info { "Running..." }
        @threads = []
        @queue_handler.poll do |messages, stats|
          messages = [ messages ] unless messages.is_a?(Array)
          messages.each do |msg|
            break if @should_stop
            break if @thread_limit && @threads.length >= @thread_limit
            logger.info { "Processing Message #{msg.message_id}... "}
            logger.debug { "Message: #{msg}" }

            message_handler = SQSMessageHandler.new(logger: logger,
                                                    queue_handler: @queue_handler.dup,
                                                    message: msg.dup,
                                                    event_handler: @event_handler.dup)
            t = Thread.new(message_handler) { |mh| mh.process }
            @threads << t
          end
          @threads.keep_if(&:alive?)
          throw :stop_polling if @should_stop
          throw :skip_delete
        end

        @running = false
      end

      def stop
        @should_stop = true
      end

      # Agent
    end

    # Restore
  end

end
